# -*- coding: utf-8 -*-
import logging
import math
import Speaker
import Segment
import IntersegmentDistanceTable
from ACCDIST.vowel_distance import VowelComparisonTable, VowelPairDistance, Transcription, VowelPair
import IntersegmentDistanceTable

class AccentTranscriptionTable:
    def __init__(self, model):
        self.table = dict()
        for accent in model.accents:
            self.table[accent] = dict()
            for transcription in model.transcriptions:
                self.table[accent][transcription] = None

    def __getitem__(self, item):
        return self.table[item]


class MeanDistanceTable:
    def __init__(self, model=None, table=None):
        if model and table:
            raise Exception("Choose one")
        if not table and not model:
            raise Exception("Choose one")
        if table:
            self.t = table
        elif model:
            self.t = dict()
            self.accent_table = dict()
            self.accent_table_for_R = dict()
            self.segment_table = dict()
            data_sum, count = self.sum_segments(model, model.speakers_with_segments(5))
            for accent in model.accents:
                self.t[accent] = self.calculate_mean(count, data_sum, model.transcriptions, accent)
                self.accent_table[accent] = self.calculate_accent_table(accent)
                #self.segment_table[accent] = self.calculate_segment_table(accent)
            # t is a dictionary of accent: segments, where each segment has the mean value for data (MFCCs)

    def __getitem__(self, item):
        return self.t[item]

    def calculate_mean(self, count, data_sum, transcriptions, accent):
        segments = []
        speaker = Speaker(accent, accent, "")
        # logging.error("Building mean distance table for accent %s" % accent)
        for transcription in transcriptions:
            if data_sum[accent][transcription] is not None:
                # logging.error("Building mean distance table for transcription %s" % transcription)
                mean_data = [x / count[accent][transcription] for x in data_sum[accent][transcription]]
                segments.append(Segment(speaker, transcription, mean_data))
        return segments

    def sum_segments(self, model, speakers):
        data_sum = AccentTranscriptionTable(model)
        count = AccentTranscriptionTable(model)
        for speaker in speakers:
            accent = speaker.accent
            # logging.error("Building mean distance table for speaker %s" % speaker.id)
            for segment in model.get_speakers_segments(speaker):
                if data_sum[accent][segment.transcription] is None:
                    data_sum[accent][segment.transcription] = [0.0 for i in range(model.data_length)]
                if count[accent][segment.transcription] is None:
                    count[accent][segment.transcription] = 0
                data_sum[accent][segment.transcription] = [sum(x) for x in zip(data_sum[accent][segment.transcription],
                                                                               segment.data)]  # For each segment with matching accent and transcription, add data points to data_sum.
                count[accent][segment.transcription] += 1
        return data_sum, count

    def exclude_speaker(self, model, speaker):
        new_mean_table = self.t.copy()
        data_sum, count = self.sum_segments(model, [s for s in model.speakers if s is not speaker])
        new_mean_table[speaker.accent] = self.calculate_mean(count, data_sum, model.transcriptions, speaker.accent)
        return MeanDistanceTable(table=new_mean_table)

    def speaker_list(self):
        mean_speaker_table = []
        for accent in self.t:
            # If there is any data for the relevant accent, e.g. once 'self' has been eliminated
            if len(self.t[accent]) > 0:
                speaker = self.t[accent][0].speaker  # get mean speaker (identical for all segments, use first)
                speaker.set_segments(self[accent])
                speaker.set_intersegment_distance_table()
                mean_speaker_table.append(speaker)
        return mean_speaker_table

    def calculate_accent_table(self, accent):
        accent_table = VowelComparisonTable([segment.transcription for segment in self.t[accent]])

        for segment1 in self.t[accent]:
            for segment2 in self.t[accent]:
                accent_table[segment1.transcription][segment2.transcription] = IntersegmentDistanceTable.compute_segment_distance(segment1, segment2)

        return accent_table

    def calculate_segment_table(self, accent):
        original_table = self.accent_table[accent].table
        segment_table = VowelComparisonTable([Transcription(x).vowel for x in original_table.keys()])

        # Find all unique vowel combinations (e.g. (eh, y), (y, eh), etc.)
        # and all vowel pair distances from all transcriptions : eh (from paneh), y, 1.7), (eh (from tehpah), y, 2.4)...
        pair_distances = []
        vowel_combinations = set()
        for row, columns in original_table.items():
            for column, value in columns.items():
                v = VowelPairDistance(row, column, value)
                vowel_combinations.add(VowelPair(v.vowel1, v.vowel2))
                pair_distances.append(v)

        # Calculate means: for all pairs involving (eh, y), mean distance.
        for combo in vowel_combinations:
            matching = list(i for i in pair_distances if i.vowel1 == combo.vowel1 and i.vowel2 == combo.vowel2)
            total = sum(x.value for x in matching)
            mean = total / len(matching)
            segment_table[combo.vowel1][combo.vowel2] = mean

        return segment_table

# /* compute absolute difference between tables */
def compute_absolute_table_distance(segments1, segments2):
    dist=0.0
    count=0.0

    for segment1 in segments1:
        for segment2 in segments2:
            if segment1.transcription == segment2.transcription:
                dist += IntersegmentDistanceTable.compute_segment_distance(segment1, segment2)
                count += 1.0
    if dist == 0:
        logging.error("dist is 0 for %s and %s" % ([s.transcription for s in segments1], [s.transcription for s in segments2]))
        return None
    if count == 0:
        logging.error("count is 0 for segments of length %d and %d" % len(segments1), len(segments2))
        return None
    return dist/count


def compute_correlation_table_distance(speaker_x, speaker_y):
    count = 0.0
    sumx = 0.0
    sumy  = 0.0
    sumxx = 0.0
    sumyy = 0.0
    sumxy = 0.0
    segments = set([s1.transcription for s1 in speaker_x.segments]).intersection(set([s2.transcription for s2 in speaker_y.segments]))

    for segment_a in segments:
        for segment_b in segments:
            distance_x = speaker_x.intersegment_distance_table.get(segment_a, segment_b)
            distance_y = speaker_y.intersegment_distance_table.get(segment_a, segment_b)
            sumx += distance_x
            sumy += distance_y
            sumxx += distance_x * distance_x
            sumyy += distance_y * distance_y
            sumxy += distance_x * distance_y
            count += 1

    if count == 0:
        logging.error("correlation analysis: no matching elements")
        logging.error("Speaker1 = %s, Speaker2 = %s" % speaker_x, speaker_y)
    #print("{}{}".format(speaker_x, speaker_y))
    return (count*sumxy-sumx*sumy)/(math.sqrt(count*sumxx-sumx*sumx)*math.sqrt(count*sumyy-sumy*sumy))
    # Pearson's Correlation Coefficient for the 2 speakers is N(sum)xy - (sum)x(sum)y / root(N(sum)x^2-(sum)x(sum)x)root(N(sum)y^2-(sum)y(sum)y)

def absolute_distance_with_mean(model, meanAccentTable):
    correct = 0
    bad = 0
    for speaker in model.speakers_with_segments(5):
        minimum_distance = 1E10
        closest_accent = 0
        for accent in model.accents:
            d = compute_absolute_table_distance(speaker.segments,
                                                             meanAccentTable[accent])
            model.absolute_mean_all_table[speaker.id][accent] = d
            if d is not None and d < minimum_distance:
                minimum_distance = d
                closest_accent = accent

        logging.info("Speaker \t%s\t is closest to accent \t%s\t with a distance of \t%s\t" % (
            speaker.id, closest_accent, minimum_distance))
        if speaker.accent == closest_accent:
            correct += 1
        else:
            bad += 1
        model.absolute_mean_closest_table[speaker.accent][closest_accent] += 1

def absolute_distance_with_mean_excluding_self(model, meanAccentTable):
    correct = 0
    bad = 0
    for speaker in model.speakers_with_segments(5):
        minimum_distance = 1E10
        closest_accent = 0
        selfexcluding_mean_accent_table = meanAccentTable.exclude_speaker(model, speaker)
        for accent in model.accents:
            d = compute_absolute_table_distance(speaker.segments, selfexcluding_mean_accent_table[accent])
            model.absolute_selfexcluding_mean_all_table[speaker.id][accent] = d
            if d is not None and d < minimum_distance:
                minimum_distance = d
                closest_accent = accent

        logging.info("Speaker \t%s\t is closest to accent \t%s\t with a distance of \t%s\t" % (
            speaker.id, closest_accent, minimum_distance))
        if speaker.accent == closest_accent:
            correct += 1
        else:
            bad += 1
        model.absolute_selfexcluding_mean_closest[speaker.accent][closest_accent] += 1

def correlation_with_mean(model, mean_speaker_table):
    correct = 0
    bad = 0
    for individual_speaker in model.speakers_with_segments(5):
        maximum_correlation = -1
        closest_accent = 0
        for mean_speaker in mean_speaker_table:
            c = compute_correlation_table_distance(individual_speaker, mean_speaker)
            model.accdist_mean_all_table[individual_speaker.id][mean_speaker.id] = c
            if c is not None and c > maximum_correlation:
                maximum_correlation = c
                closest_accent = mean_speaker.accent

        logging.info("Speaker \t%s\t is closest to accent \t%s\t with a correlation of \t%s\t" % (
                individual_speaker.id, closest_accent, maximum_correlation))
        if individual_speaker.accent == closest_accent:
            correct += 1
        else:
            bad += 1
        model.accdist_mean_closest_table[individual_speaker.accent][closest_accent] += 1


def correlation_with_mean_excluding_self(model, meanDistanceTable):
    correct = 0
    bad = 0
    for individual_speaker in model.speakers_with_segments(5):
        maximum_correlation = -1
        closest_accent = 0
        mean_speaker_list = meanDistanceTable.exclude_speaker(model, individual_speaker).speaker_list()
        for mean_speaker in mean_speaker_list:
            c = compute_correlation_table_distance(individual_speaker, mean_speaker)
            model.accdist_selfexcluding_mean_all_table[individual_speaker.id][mean_speaker.id] = c
            if c is not None and c > maximum_correlation:
                maximum_correlation = c
                closest_accent = mean_speaker.accent

        logging.info("Speaker \t%s\t is closest to accent \t%s\t with a correlation of \t%s\t" % (
                individual_speaker.id, closest_accent, maximum_correlation))
        if individual_speaker.accent == closest_accent:
            correct += 1
        else:
            bad += 1
        model.accdist_selfexcluding_mean_closest[individual_speaker.accent][closest_accent] += 1